/*
 *  EXECUTER, {akii-root}/src/executer.c
 *
 * ==========================================================================
 *
 *                      Arena Kliens II. Releases
 *
 * ==========================================================================
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * A legkiralyabb szovegszerkeszto: VIM - Vi IMproved 5  <http://www.vim.org>
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "interpreter.h"
#include "variables.h"
#include "error.h"
#include "debug.h"

int do_alias(struct parancs *par)
{
    char *p;
    struct aliasstruct *as;
    struct aliasstr *ali;
    struct valtozo *var;
    int i;
    char temp[256];
    int save_varflag = varflag;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_alias, "?");
#endif
    as = par->ertek;
    p = as->nev;
    varflag = VFLAG_NORMAL;
    var = getvar(&varlist, &p, 1, 0);
    varflag = save_varflag;
    if ((!var) || (var->tipus != VAR_NORMAL) || (var->ertek[0] != 0)) {
	if (!((var->tipus == VAR_ALIAS) && (!novice))) {
	    errno = EALIASNEM;
	    strcpy(errtemp, as->nev);
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_alias);
#endif
	    return (errno);
	}
    }
    if (var->tipus != VAR_ALIAS) {
	var->tipus = VAR_ALIAS;
	free(var->ertek);
    }
    else {
	ali = (struct aliasstr *) (var->ertek);
	freepar(ali->par);
	for (i = 0; i < ali->cpnum; i++)
	    free(ali->cps[i]);
	if (ali->cps)
	    free(ali->cps);
	free(ali);
    }
    ali = malloc(sizeof(struct aliasstr));

    ali->cpnum = as->cpnum;
    ali->par = as->par;
    ali->par->refnum++;
    ali->cps = NULL;
    if (ali->cpnum) {
	ali->cps = malloc(ali->cpnum * sizeof(char *));

	for (i = 0; i < ali->cpnum; i++) {
	    ali->cps[i] = malloc(strlen(as->cps[i] + 1));
	    strcpy(ali->cps[i], as->cps[i]);
	}
    }
    var->ertek = (char *) ali;
#ifdef AKII_DEBUG
    dbgleave(dbg_do_alias);
#endif
    return (0);
}
int igaze(double d, char *s)
{
    return ((d != 0) || (istrcmp(s, "ON") == 0) || (istrcmp(s, "TRUE") == 0) || (istrcmp(s, "IGAZ") == 0) || (istrcmp(s, "BE") == 0));
}
int do_checkpoint(struct parancs *par)
{
    int i;
    char *s;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_checkpoint, "?");
#endif
    i = doparancssor(par->ertek);
    if (i == DO_ESCAPE) {
	if (!--checkpoint_level)
	    i = 0;
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_do_checkpoint);
#endif
    return (i);
}
int do_if(struct parancs *par)
{
    struct ifstruct *is;
    double i;
    char *s;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_if, "?");
#endif
    is = par->ertek;
    s = dokif(is->kif);
    if (!s) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_if);
#endif
	return (1);
    }
    i = strtod(s, NULL);
    if (errno == EINVAL)
	errno = 0;
    if (igaze(i, s)) {
	free(s);
#ifdef AKII_DEBUG
	dbgleave(dbg_do_if);
#endif
	return (doparancssor(is->par));
    }
    free(s);
    if (is->elsepar) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_if);
#endif
	return (doparancssor(is->elsepar));
    }
    else {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_if);
#endif
	return (0);
    }
}
int do_return(struct parancs *par)
{
    struct kif *kif;
    char *s, temp[] = "akii.RETURN", *t = temp;
    struct valtozo *var;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_return, "?");
#endif
    if (!par->ertek) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_return);
#endif
	return (DO_RETURN);
    }
    var = getvar(&bivarlist, &t, 0, 0);
    if (!var) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_return);
#endif
	return (DO_RETURN);
    }
    s = dokif(par->ertek);
    free(var->ertek);
    var->ertek = s;
#ifdef AKII_DEBUG
    dbgleave(dbg_do_return);
#endif
    return (DO_RETURN);
}
int do_do(struct parancs *par)
{
    struct wstruct *ws;
    int i;
    double f;
    char *s;

    ws = par->ertek;
#ifdef AKII_DEBUG
    dbgenter(dbg_do_do, "?");
#endif
    while (1) {
	i = doparancssor(ws->par);
	if (i == DO_ESCAPE)
	    break;
	if ((i == DO_NOMOREON) || (i == DO_NEXTON))
	    break;
	if (i == DO_BREAK) {
	    brkcnt_level--;
	    break;
	}
	if (i == DO_CONTINUE) {
	    brkcnt_level--;
	    if (brkcnt_level)
		break;
	}
	if (i == DO_RETURN) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_do);
#endif
	    return (DO_RETURN);
	}
	if (i > 0) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_do);
#endif
	    return (i);
	}
	s = dokif(ws->kif);
	if (!s) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_do);
#endif
	    return (1);
	}
	f = strtod(s, NULL);
	if (errno == EINVAL)
	    errno = 0;
	if (igaze(f, s))
	    free(s);
	else {
	    free(s);
	    break;
	}
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_do_do);
#endif
    if (brkcnt_level || checkpoint_level)
	return (i);
    return (0);
}
int do_while(struct parancs *par)
{
    struct wstruct *ws;
    int i;
    double f;
    char *s;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_while, "?");
#endif
    ws = par->ertek;
    while (1) {
	s = dokif(ws->kif);
	if (!s) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_while);
#endif
	    return (1);
	}
	f = strtod(s, NULL);
	if (errno == EINVAL)
	    errno = 0;
	if (igaze(f, s)) {
	    free(s);
	    i = doparancssor(ws->par);
	    if (i == DO_ESCAPE)
		break;
	    if ((i == DO_NOMOREON) || (i == DO_NEXTON))
		break;
	    if (i == DO_BREAK) {
		brkcnt_level--;
		break;
	    }
	    if (i == DO_CONTINUE) {
		if (--brkcnt_level)
		    break;
	    }
	    if (i == DO_RETURN) {
#ifdef AKII_DEBUG
		dbgleave(dbg_do_while);
#endif
		return (DO_RETURN);
	    }
	    if (i > 0) {
#ifdef AKII_DEBUG
		dbgleave(dbg_do_while);
#endif
		return (i);
	    }
	}
	else {
	    free(s);
	    break;
	}
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_do_while);
#endif
    if (brkcnt_level || checkpoint_level)
	return (i);
    return (0);
}
int do_for(struct parancs *par)
{
    struct forstruct *fs;
    int i;
    double f;
    char *s;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_for, "?");
#endif
    fs = par->ertek;
    s = dokif(fs->kifek[0]);
    if (!s) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_for);
#endif
	return (1);
    }
    free(s);
    while (1) {
	s = dokif(fs->kifek[1]);
	if (!s) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_for);
#endif
	    return (1);
	}
	f = strtod(s, NULL);
	if (errno == EINVAL)
	    errno = 0;
	if (igaze(f, s)) {
	    free(s);
	    i = doparancssor(fs->par);
	    if (i == DO_ESCAPE)
		break;
	    if ((i == DO_NOMOREON) || (i == DO_NEXTON))
		break;
	    if (i == DO_BREAK) {
		brkcnt_level--;
		break;
	    }
	    if (i == DO_CONTINUE) {
		if (--brkcnt_level)
		    break;
	    }
	    if (i == DO_RETURN) {
#ifdef AKII_DEBUG
		dbgleave(dbg_do_for);
#endif
		return (DO_RETURN);
	    }
	    if (i > 0) {
#ifdef AKII_DEBUG
		dbgleave(dbg_do_for);
#endif
		return (i);
	    }
	}
	else {
	    free(s);
	    break;
	}
	s = dokif(fs->kifek[2]);
	if (!s) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_for);
#endif
	    return (1);
	}
	free(s);
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_do_for);
#endif
    if (brkcnt_level || checkpoint_level)
	return (i);
    return (0);
}
int do_foreach(struct parancs *par)
{
    struct foreachstruct *fes;
    struct valtozo *var, *ivar, *var2, *jvar;
    int i;
    char *s, *j, *jj;
    int list = 0;
    int save_varflag = varflag;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_foreach, "?");
#endif
    fes = par->ertek;
    s = fes->tomb;
    if (strcmp(s, ".") == 0) {
	var = varlist;
	list = 1;
    }
    else
	var = getvar(&varlist, &s, 0, 1);
    if (errno || (!var)) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_foreach);
#endif
	return (1);
    }
    if (!list)
	if (var->tipus != VAR_STRUCT) {
	    errno = ENONSTRUCT;
	    strcpy(errtemp, var->nev);
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_foreach);
#endif
	    return (1);
	}
    s = fes->inev;
    jj = j = strchr(s, ':');
    if (j) {
	*j = 0;
	j++;
    }
    ivar = getvar(&localvars[locallevel], &s, 0, 0);
    if (!ivar) {
	s = fes->inev;
	varflag = VFLAG_LOCAL;
	ivar = getvar(&localvars[locallevel], &s, 1, 0);
	varflag = save_varflag;
    }
    else {
	errno = EFOGLALT;
	strcpy(errtemp, fes->inev);
	if (jj)
	    *jj = ':';
#ifdef AKII_DEBUG
	dbgleave(dbg_do_foreach);
#endif
	return (1);
    }
    if (j) {
	jvar = getvar(&localvars[locallevel], &j, 0, 0);
	if (!jvar) {
	    j = fes->inev + strlen(fes->inev) + 1;
	    varflag = VFLAG_LOCAL;
	    jvar = getvar(&localvars[locallevel], &j, 1, 0);
	    varflag = save_varflag;
	}
	else {
	    errno = EFOGLALT;
	    strcpy(errtemp, fes->inev + strlen(fes->inev) + 1);
	    if (jj)
		*jj = ':';
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_foreach);
#endif
	    return (1);
	}
    }
    if (errno) {
	if (jj)
	    *jj = ':';
#ifdef AKII_DEBUG
	dbgleave(dbg_do_foreach);
#endif
	return (1);
    }
    ivar->tipus = VAR_FOREACH;
    free(ivar->ertek);
    ivar->ertek = NULL;
    ivar->notdeletable++;
    if (j)
	jvar->notdeletable++;
    if (!list)
	var2 = (struct valtozo *) var->ertek;
    else
	var2 = var;
    var->notdeletable++;
    while (var2) {
	ivar->ertek = (char *) var2;
	if (j) {
	    free(jvar->ertek);
	    jvar->ertek = strdup(var2->nev);
	}
	i = doparancssor(fes->par);
	if (i == DO_ESCAPE)
	    break;
	if ((i == DO_NOMOREON) || (i == DO_NEXTON))
	    break;
	if (i == DO_BREAK) {
	    brkcnt_level--;
	    break;
	}
	if (i == DO_CONTINUE) {
	    if (--brkcnt_level)
		break;
	}
	if (i == DO_RETURN)
	    break;
	if (i > 0) {
	    var->notdeletable--;
	    ivar->notdeletable--;
	    if (j)
		jvar->notdeletable--;
	    freevarlist(ivar, 0);
	    if (j)
		freevarlist(jvar, 0);
	    if (jj)
		*jj = ':';
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_foreach);
#endif
	    return (i);
	}
	var2 = var2->next;
    }
    var->notdeletable--;
    ivar->notdeletable--;
    if (j)
	jvar->notdeletable--;
    freevarlist(ivar, 0);
    if (j)
	freevarlist(jvar, 0);
    if (jj)
	*jj = ':';
    if (i == DO_RETURN) {
#ifdef AKII_DEBUG
	dbgleave(dbg_do_foreach);
#endif
	return (DO_RETURN);
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_do_foreach);
#endif
    if (brkcnt_level || checkpoint_level)
	return (i);
    return (0);
}
int do_on(struct parancs *par)
{
    struct onstruct *ons;
    struct onstr *on;
    struct onstr *p, *w;
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_on, "?");
#endif
    on = malloc(sizeof(struct onstr));

    ons = par->ertek;
    if (ons->nev) {
	on->nev = malloc(strlen(ons->nev) + 1);
	strcpy(on->nev, ons->nev);
    }
    else
	on->nev = ons->nev;
    on->tipus = ons->tipus;
    on->tiltott = 0;
    on->szint = ons->szint;
    on->next = on->prev = NULL;
    on->kifek = ons->kifek;
    if (ons->regkifek) {
	on->regkifek = malloc(sizeof(char *) * on->kifek);

	for (i = 0; i < on->kifek; i++) {
	    on->regkifek[i] = malloc(strlen(ons->regkifek[i]) + 1);
	    strcpy(on->regkifek[i], ons->regkifek[i]);
	}
    }
    else
	on->regkifek = NULL;
    if (ons->par)
	ons->par->refnum++;
    on->par = ons->par;
    on->justmark = 0;
    if (par->tipus == PAR_ON)
	p = onlist;
    else
	p = onsendlist;
    if (!p) {
	if (par->tipus == PAR_ON)
	    onlist = on;
	else
	    onsendlist = on;
#ifdef AKII_DEBUG
	dbgleave(dbg_do_on);
#endif
	return (0);
    }
    w = p;
    if (on->nev)
	while (w) {
	    if (w->nev)
		if (istrcmp(w->nev, on->nev) == 0) {
		    if (novice) {
			errno = EONNEVVAN;
			strcpy(errtemp, on->nev);
			free(on);
#ifdef AKII_DEBUG
			dbgleave(dbg_do_on);
#endif
			return (errno);
		    }
		    else {
			freeoneon(w, (par->tipus == PAR_ONSEND));
			break;
		    }
		}
	    w = w->next;
	}
    if (par->tipus == PAR_ON)
	p = onlist;
    else
	p = onsendlist;
    if (!p) {
	if (par->tipus == PAR_ON)
	    onlist = on;
	else
	    onsendlist = on;
#ifdef AKII_DEBUG
	dbgleave(dbg_do_on);
#endif
	return (0);
    }
    while (p->szint > on->szint) {
	if (!p->next)
	    break;
	p = p->next;
    }
    if (p->szint <= on->szint) {
	on->prev = p->prev;
	p->prev = on;
	on->next = p;
	if (!on->prev) {
	    if (par->tipus == PAR_ON)
		onlist = on;
	    else
		onsendlist = on;
	}
	else
	    on->prev->next = on;
#ifdef AKII_DEBUG
	dbgleave(dbg_do_on);
#endif
	return (0);
    }
    if (p->szint > on->szint) {
	p->next = on;
	on->prev = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_do_on);
#endif
	return (0);
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_do_on);
#endif
    return (0);
}
int do_timer(struct parancs *par)
{
    struct timerstr *timer, *p = timerlist, *w;
    struct timerstruct *t;
    int i = 0;
    char *s;
    struct valtozo *var;

#ifdef AKII_DEBUG
    dbgenter(dbg_do_timer, "?");
#endif
    t = par->ertek;
    timer = malloc(sizeof(struct timerstr));

    if (t->nev) {
	if (t->nev[0] == '$') {
	    s = t->nev + 1;
	    var = getvar(&varlist, &s, 0, 0);
	    if (!var || var->tipus != VAR_NORMAL) {
		errno = EWRONGTIMER;
		strcpy(errtemp, t->nev);
		free(timer);
#ifdef AKII_DEBUG
		dbgleave(dbg_do_timer);
#endif
		return (errno);
	    }
	    s = var->ertek;
	}
	else
	    s = t->nev;
	timer->nev = malloc(strlen(s) + 1);
	strcpy(timer->nev, s);
    }
    else
	timer->nev = t->nev;
    if (t->tipus == '1' || t->tipus == '2') {
	s = t->ido;
	var = getvar(&varlist, &s, 0, 0);
	if (!var || var->tipus != VAR_NORMAL) {
	    errno = EWRONGTIMER;
	    strcpy(errtemp, t->nev);
	    free(timer);
#ifdef AKII_DEBUG
	    dbgleave(dbg_do_timer);
#endif
	    return (errno);
	}
	timer->ido = strtod(var->ertek, NULL);
	if (t->tipus == '1')
	    timer->tipus = '~';
	else
	    timer->tipus = '!';
    }
    else {
	timer->ido = (int) t->ido;
	timer->tipus = t->tipus;
    }
    timer->tiltott = 0;
    timer->last = time(NULL);
    timer->next = timer->prev = NULL;
    t->par->refnum++;
    timer->par = t->par;
    if (!timerlist) {
	timerlist = timer;
#ifdef AKII_DEBUG
	dbgleave(dbg_do_timer);
#endif
	return (0);
    }
    if (timer->nev)
	while (p) {
	    w = p;
	    if (p->nev)
		if ((i = istrcmp(p->nev, timer->nev)) == 0) {
		    errno = ETIMERVAN;
		    strcpy(errtemp, p->nev);
		    free(timer);
#ifdef AKII_DEBUG
		    dbgleave(dbg_do_timer);
#endif
		    return (errno);
		}
	    if (i > 0)
		break;
	    p = p->next;
	}
    if (p) {
	timer->prev = p->prev;
	timer->next = p;
	if (p->prev)
	    p->prev->next = timer;
	else {
	    timerlist = timer;
	}
	p->prev = timer;
#ifdef AKII_DEBUG
	dbgleave(dbg_do_timer);
#endif
	return (0);
    }
    w->next = timer;
    timer->prev = w;
#ifdef AKII_DEBUG
    dbgleave(dbg_do_timer);
#endif
    return (0);
}
int dooneparancs(struct parancs *par)
{
    char *p;

#ifdef AKII_DEBUG
    dbgenter(dbg_dooneparancs, "?");
#endif
    if (!par) {
#ifdef AKII_DEBUG
	dbgleave(dbg_dooneparancs);
#endif
	return (0);
    }
    switch (par->tipus) {
	case PAR_KIF:
	    p = dokif(par->ertek);
	    if (p)
		free(p);
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (errno);
	case PAR_BREAK:
	    brkcnt_level = (int) par->ertek;
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (DO_BREAK);
	case PAR_ESCAPE:
	    checkpoint_level = (int) par->ertek;
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (DO_ESCAPE);
	case PAR_NOMOREON:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (DO_NOMOREON);
	case PAR_CONTINUE:
	    brkcnt_level = (int) par->ertek;
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (DO_CONTINUE);
	case PAR_NEXTON:
	    nexton_level = (int) par->ertek;
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (DO_NEXTON);
	case PAR_RETURN:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_return(par));
	case PAR_ALIAS:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_alias(par));
	case PAR_IF:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_if(par));
	case PAR_WHILE:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_while(par));
	case PAR_DO:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_do(par));
	case PAR_FOR:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_for(par));
	case PAR_FOREACH:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_foreach(par));
	case PAR_ON:
	case PAR_ONSEND:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_on(par));
	case PAR_TIMER:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_timer(par));
	case PAR_CHECKPOINT:
#ifdef AKII_DEBUG
	    dbgleave(dbg_dooneparancs);
#endif
	    return (do_checkpoint(par));
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_dooneparancs);
#endif
    return (1);
}
int doparancssor(struct parancs *par)
{
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_doparancssor, "?");
#endif
    while (par) {
	i = dooneparancs(par);
	if (errno)
	    break;
	if ((i == DO_CONTINUE) || (i == DO_BREAK) || (i == DO_ESCAPE) || (i == DO_NEXTON) || (i == DO_NOMOREON))
	    break;
	if (i == DO_RETURN)
	    break;
	par = par->next;
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_doparancssor);
#endif
    return (i);
}
